/*
 * Routines to decode user commands.
 *
 * This is all table driven.
 * A command table is a sequence of command descriptors.
 * Each command descriptor is a sequence of bytes with the following format:
 *	<c1><c2>...<cN><0><action>
 * The characters c1,c2,...,cN are the command string; that is,
 * the characters which the user must type.
 * It is terminated by a null <0> byte.
 * The byte after the null byte is the action code associated
 * with the command string.
 *
 * There may be many command tables.
 * The first (default) table is built-in.
 * Other tables are read in from "lesskey" files.
 * All the tables are linked together and are searched in order.
 */

#include "less.h"
#include "cmd.h"

#ifdef MSDOS
#include <io.h>
#include <fcntl.h>
#endif

#define	ESC	CONTROL('[')

/*
 * Command table is ordered roughly according to expected
 * frequency of use, so the common commands are near the beginning.
 */
static char cmdtable[] =
{
	'\r',0,				A_F_LINE,
	'\n',0,				A_F_LINE,
	'e',0,				A_F_LINE,
	'j',0,				A_F_LINE,
	CONTROL('E'),0,			A_F_LINE,
	CONTROL('N'),0,			A_F_LINE,
	'k',0,				A_B_LINE,
	'y',0,				A_B_LINE,
	CONTROL('Y'),0,			A_B_LINE,
	CONTROL('K'),0,			A_B_LINE,
	CONTROL('P'),0,			A_B_LINE,
	'd',0,				A_F_SCROLL,
	CONTROL('D'),0,			A_F_SCROLL,
	'u',0,				A_B_SCROLL,
	CONTROL('U'),0,			A_B_SCROLL,
	' ',0,				A_F_SCREEN,
	'f',0,				A_F_SCREEN,
	CONTROL('F'),0,			A_F_SCREEN,
	CONTROL('V'),0,			A_F_SCREEN,
	'b',0,				A_B_SCREEN,
	CONTROL('B'),0,			A_B_SCREEN,
	ESC,'v',0,			A_B_SCREEN,
	'z',0,				A_F_WINDOW,
	'w',0,				A_B_WINDOW,
	'R',0,				A_FREPAINT,
	'r',0,				A_REPAINT,
	CONTROL('R'),0,			A_REPAINT,
	CONTROL('L'),0,			A_REPAINT,
	'g',0,				A_GOLINE,
	'<',0,				A_GOLINE,
	ESC,'<',0,			A_GOLINE,
	'p',0,				A_PERCENT,
	'%',0,				A_ANYBRAC,
	'{',0,				A_BRACLC,
	'}',0,				A_BRACRC,
	'(',0,				A_BRACLP,
	')',0,				A_BRACRP,
	'[',0,				A_BRACLS,
	']',0,				A_BRACRS,
	'G',0,				A_GOEND,
	ESC,'>',0,			A_GOEND,
	'>',0,				A_GOEND,

	'0',0,				A_DIGIT,
	'1',0,				A_DIGIT,
	'2',0,				A_DIGIT,
	'3',0,				A_DIGIT,
	'4',0,				A_DIGIT,
	'5',0,				A_DIGIT,
	'6',0,				A_DIGIT,
	'7',0,				A_DIGIT,
	'8',0,				A_DIGIT,
	'9',0,				A_DIGIT,

	'=',0,				A_STAT,
	CONTROL('G'),0,			A_STAT,
	':','f',0,			A_STAT,
	'/',0,				A_F_SEARCH,
	'?',0,				A_B_SEARCH,
	ESC,'/',0,			A_T_SEARCH,
	'n',0,				A_AGAIN_SEARCH,
	ESC,'n',0,			A_REVERSE_SEARCH,
	'm',0,				A_SETMARK,
	'\'',0,				A_GOMARK,
	CONTROL('X'),CONTROL('X'),0,	A_GOMARK,
	'E',0,				A_EXAMINE,
	':','e',0,			A_EXAMINE,
	CONTROL('X'),CONTROL('V'),0,	A_EXAMINE,
	'N',0,				A_NEXT_FILE,
	'P',0,				A_PREV_FILE,
	':','n',0,			A_NEXT_FILE,
	':','p',0,			A_PREV_FILE,
	'-',0,				A_TOGGLE_OPTION,
	':','t','a',0,			A_1_TOGGLE_OPTION|('t'),
	's',0,				A_1_TOGGLE_OPTION|('l'),
	'_',0,				A_DISP_OPTION,
	'v',0,				A_VISUAL,
	'!',0,				A_SHELL,
	'+',0,				A_FIRSTCMD,

	'H',0,				A_HELP,
	'h',0,				A_HELP,
	'V',0,				A_VERSION,
	'q',0,				A_QUIT,
	':','q',0,			A_QUIT,
	':','Q',0,			A_QUIT,
	'Z','Z',0,			A_QUIT
};

/*
 * Structure to support a list of command tables.
 */
struct tablelist
{
	struct tablelist *t_next;
	char *t_start;
	char *t_end;
};

/*
 * Structure for the default command table.
 */
static struct tablelist deftable =
	{ NULL, cmdtable, cmdtable+sizeof(cmdtable) };

/*
 * List of tables; initially contains only the default table.
 */
static struct tablelist *tables = &deftable;

/*
 * Buffer to hold current command string.
 */
static char kbuf[MAX_CMDLEN+1];
static char *kp = kbuf;

static int cmd_search();

/*
 * Decode a command character and return the associated action.
 */
	public int
cmd_decode(c)
	int c;
{
	register struct tablelist *t;
	register int action;

	/*
	 * Append the new command character to the command string in kbuf.
	 */
	*kp++ = c;
	*kp = '\0';

	/*
	 * Search thru all the command tables.
	 * Stop when we find an action which is not A_INVALID.
	 */
	for (t = tables;  t != NULL;  t = t->t_next)
	{
		action = cmd_search(t->t_start, t->t_end);
		if (action != A_INVALID)
			break;
	}

	if (action != A_PREFIX)
		/*
		 * This is not a prefix character.
		 */
		noprefix();

	return (action);
}

/*
 * Indicate that we're not in a prefix command
 * by resetting the command buffer pointer.
 */
	public void
noprefix()
{
	kp = kbuf;
}

/*
 * Search a command table for the current command string (in kbuf).
 */
	static int
cmd_search(table, endtable)
	char *table;
	char *endtable;
{
	register char *p;
	register char *q;

	for (p = table, q = kbuf;  p < endtable;  p++, q++)
	{
		if (*p == *q)
		{
			/*
			 * Current characters match.
			 * If we're at the end of the string, we've found it.
			 * Return the action code, which is the character
			 * after the null at the end of the string
			 * in the command table.
			 */
			if (*p == '\0')
				return (p[1] & 0377);
		} else if (*q == '\0')
		{
			/*
			 * Hit the end of the user's command,
			 * but not the end of the string in the command table.
			 * The user's command is incomplete.
			 */
			return (A_PREFIX);
		} else
		{
			/*
			 * Not a match.
			 * Skip ahead to the next command in the
			 * command table, and reset the pointer
			 * to the user's command.
			 */
			while (*p++ != '\0') ;
			q = kbuf-1;
		}
	}
	/*
	 * No match found in the entire command table.
	 */
	return (A_INVALID);
}

#if USERFILE
/*
 * Set up a user command table, based on a "lesskey" file.
 */
	public int
add_cmdtable(filename)
	char *filename;
{
	register struct tablelist *t;
	register POSITION len;
	register long n;
	register int f;

	/*
	 * Try to open the lesskey file.
	 * If we can't, return an error.
	 */
        f = open(filename, O_RDONLY | O_BINARY);
	if (f < 0)
		return (-1);

	/*
	 * Read the file into the user table.
	 * We first figure out the size of the file and allocate space for it.
	 * {{ Minimal error checking is done here.
	 *    A garbage .less file will produce strange results.
	 *    To avoid a large amount of error checking code here, we
	 *    rely on the lesskey program to generate a good .less file. }}
	 */
	len = filesize(f);
	if (len == NULL_POSITION || len < 3)
	{
		/*
		 * Bad file (valid file must have at least 3 chars).
		 */
		close(f);
		return (-1);
	}
	if ((t = (struct tablelist *)
			calloc(1, sizeof(struct tablelist))) == NULL)
	{
		close(f);
		return (-1);
	}
        if (((t->t_start) = (char *) calloc(1, len)) == NULL)
	{
		free((char *)t);
		close(f);
		return (-1);
	}
	if (lseek(f, (offset_t)0, 0) == BAD_LSEEK)
	{
		free(t->t_start);
		free((char *)t);
		close(f);
		return (-1);
	}
	n = read(f, t->t_start, len);
	close(f);

	if (n != len || t->t_start[n-2] != '\0')
	{
		/*
		 * Several error cases are lumped together here:
		 * - Cannot read user file (n < 0).
		 * - User file is too short (a valid file must
		 *   have at least 3 chars: one char command string,
		 *   the terminating null byte, and the action byte).
		 * - The final entry in the user file is bad (it
		 *   doesn't have a null byte in the proper place).
		 * Many other error cases are not caught, such as
		 * invalid format in any except the last entry,
		 * invalid action codes, command strings too long, etc.
		 */
		free(t->t_start);
		free((char *)t);
		return (-1);
	}
	t->t_end = t->t_start + n;
	t->t_next = tables;
	tables = t;
	return (0);
}

/*
 * Try to add the lesskey file "$HOME/.less"
 */
	public void
add_hometable()
{
	char *homedir;
	char *filename;
	extern char *getenv();

#ifdef MSDOS
        filename = (char *) calloc(128, sizeof(char));
	if (filename == NULL)
		return;
        _searchenv("less.ini", "INIT", filename);
        if ( filename[0] == 0 )
          _searchenv("less.ini", "PATH", filename);
#else
	homedir = getenv("HOME");
	if (homedir == NULL)
		return;
	filename = (char *) calloc(strlen(homedir)+7, sizeof(char));
	if (filename == NULL)
		return;
	sprintf(filename, "%s/.less", homedir);
#endif

	/*
	 * Ignore errors.
	 */
	(void) add_cmdtable(filename);

	free(filename);
}
#endif
