/*
 * Stuff to manipulate file handles.
 * A file handle is a small integer that refers to a file name.
 * The point of all this is to avoid keeping big file names around
 * in many different places; instead we keep file HANDLES around
 * and all the names are stored away here in this module.
 * There are routines here to convert a handle to a name and vice versa.
 *
 * Also done in this module is keeping track of a file position
 * for every filename.  This is used to restore the last position
 * when we re-examine a previously examined file.
 */

#include "less.h"

struct fhandle {
	struct fhandle *h_next;
	HANDLE h_handle;
	char *h_filename;
	POSITION h_pos;
};

static HANDLE xhandle = (HANDLE)0;
static struct fhandle *anchor = NULL;

/*
 * Allocate a new handle structure and stick a filename in it.
 */
	static struct fhandle *
new_handle(filename)
	char *filename;
{
	register struct fhandle *p;

	p = (struct fhandle *) ecalloc(1, sizeof(struct fhandle));
	p->h_handle = ++xhandle;
	p->h_filename = filename;
	p->h_next = anchor;
	p->h_pos = (POSITION)0;
	anchor = p;
	return (p);
}

/*
 * Find a handle structure, given a handle.
 */
	static struct fhandle *
h_to_fh(handle)
	HANDLE handle;
{
	register struct fhandle *p;

	for (p = anchor;  p != NULL;  p = p->h_next)
		if (handle == p->h_handle)
			return (p);

	error("ERROR: h_to_fh");
	quit();
	/*NOTREACHED*/
}

/*
 * Find a handle structure, given a filename.
 */
	static struct fhandle *
fn_to_fh(filename)
	char *filename;
{
	register struct fhandle *p;

	for (p = anchor;  p != NULL;  p = p->h_next)
		if (strcmp(filename, p->h_filename) == 0)
			return (p);
	return (NULL);
}

/*
 * Get the handle associated with a filename.
 */
	public HANDLE
get_handle(filename)
	char *filename;
{
	register struct fhandle *p;

	if ((p = fn_to_fh(filename)) == NULL)
		p = new_handle(save(filename));
	return (p->h_handle);
}

/*
 * Stash away a filename.
 * The filename must be known to be STATIC;
 * this function, unlike get_handle(),
 * does not make a copy of the filename.
 */
	public void
save_handle(filename)
	char *filename;
{
	if (fn_to_fh(filename) == NULL)
		(void) new_handle(filename);
}

/*
 * Get the filename associated with a handle.
 */
	public char *
get_filename(handle)
	HANDLE handle;
{
	register struct fhandle *p;

	p = h_to_fh(handle);
	return (p->h_filename);
}

/*
 * Save the current position in a given file.
 */
	public void
store_pos(handle, pos)
	HANDLE handle;
	POSITION pos;
{
	register struct fhandle *p;

	p = h_to_fh(handle);
	p->h_pos = pos;
}

/*
 * Recall the current position in a file which may have been seen before.
 * If it hasn't been seen before, return NULL_POSITION.
 */
	public POSITION
recall_pos(handle)
	HANDLE handle;
{
	register struct fhandle *p;

	p = h_to_fh(handle);
	return (p->h_pos);
}
