/*
 * Routines to manipulate the "line buffer".
 * The line buffer holds a line of output as it is being built
 * in preparation for output to the screen.
 * We keep track of the PRINTABLE length of the line as it is being built.
 */

#include "less.h"

static char linebuf[1024];	/* Buffer which holds the current output line */
static char xlinebuf[1024];	/* Extension of linebuf to hold high bits */
static int curr;		/* Index into linebuf */
static int column;		/* Printable length, accounting for
				   backspaces, etc. */
static int is_null_line;	/* */

/*
 * A ridiculously complex state machine takes care of backspaces
 * when in BS_SPECIAL mode.  The complexity arises from the attempt
 * to deal with all cases, especially involving long lines with underlining,
 * boldfacing or whatever.  There are still some cases which will break it.
 *
 * There are four states:
 *	LN_NORMAL is the normal state (not in underline mode).
 *	LN_UNDERLINE means we are in underline mode.  We expect to get
 *		either a sequence like "_\bX" or "X\b_" to continue
 *		underline mode, or anything else to end underline mode.
 *	LN_BOLDFACE means we are in boldface mode.  We expect to get sequences
 *		like "X\bX\b...X\bX" to continue boldface mode, or anything
 *		else to end boldface mode.
 *	LN_UL_X means we are one character after LN_UNDERLINE
 *		(we have gotten the '_' in "_\bX" or the 'X' in "X\b_").
 *	LN_UL_XB means we are one character after LN_UL_X
 *		(we have gotten the backspace in "_\bX" or "X\b_";
 *		we expect one more ordinary character,
 *		which will put us back in state LN_UNDERLINE).
 *	LN_BO_X means we are one character after LN_BOLDFACE
 *		(we have gotten the 'X' in "X\bX").
 *	LN_BO_XB means we are one character after LN_BO_X
 *		(we have gotten the backspace in "X\bX";
 *		we expect one more 'X' which will put us back
 *		in LN_BOLDFACE).
 */
static int ln_state;		/* Currently in normal/underline/bold/etc mode? */
#define	LN_NORMAL	0	/* Not in underline, boldface or whatever mode */
#define	LN_UNDERLINE	1	/* In underline, need next char */
#define	LN_UL_X		2	/* In underline, got char, need \b */
#define	LN_UL_XB	3	/* In underline, got char & \b, need one more */
#define	LN_BOLDFACE	4	/* In boldface, need next char */
#define	LN_BO_X		5	/* In boldface, got char, need \b */
#define	LN_BO_XB	6	/* In boldface, got char & \b, need same char */

extern int bs_mode;
extern int tabstop;
extern int linenums;
extern int sendctl;
extern int twiddle;
extern int auto_wrap, ignaw;
extern int bo_width, be_width;
extern int ul_width, ue_width;
extern int sc_width, sc_height;

/*
 * Macros to manipulate the characters in the line buffer.
 * We encapsulate the manipulation here because of the complexity
 * of the "extended characters".  Each character is stored with
 * its low-order 8 bits in linebuf and the higher-order bits in
 * xlinebuf.  Normal characters are only 8 bits, but special
 * marker characters like BE_CHAR, etc. are wider than 8 bits.
 * In addition, the CARATBIT bit may be or-ed into any character
 * to cause it to be prefixed with a carat when printed.
 *
 *	STOC		Store a character near the current char position.
 *	MOVC		Move a character from one position to another.
 *	INC		Increment (or decrement) the current char position.
 *	READC		Read a character from near the current char position.
 *	SPACEUSED 	Returns the amount of space used in the buffer.
 *	SPACELEFT	Returns the amount of space remaining in the buffer.
 */
#define	STOC(i,c)		{ linebuf[curr+(i)] = (c); \
				  xlinebuf[curr+(i)] = ((c)>>8); }
#define	MOVC(i,j)		{ linebuf[curr+(i)] = linebuf[curr+(j)]; \
				  xlinebuf[curr+(i)] = xlinebuf[curr+(j)]; }
#define	INC(i)			{ curr += (i); }
#define	READC(i)		((linebuf[curr+(i)]&0377) | ((xlinebuf[curr+(i)]&0377)<<8))
#define	SPACEUSED		(curr)
#define	SPACELEFT		(sizeof(linebuf) - curr)


/*
 * Rewind the line buffer.
 */
	public void
prewind()
{
	curr = 0;
	column = 0;
	ln_state = LN_NORMAL;
	is_null_line = 0;
}

/*
 * Insert the line number (of the given position) into the line buffer.
 */
	public void
plinenum(pos)
	POSITION pos;
{
	register int lno;
	register int i;
	register int n;
	char buf[10];

	/*
	 * We display the line number at the start of each line
	 * only if the -N option is set.
	 */
	if (linenums != 2)
		return;

	/*
	 * Get the line number and put it in the current line.
	 * {{ Note: since find_linenum calls forw_raw_line,
	 *    it may seek in the input file, requiring the caller
	 *    of plinenum to re-seek if necessary. }}
	 */
	lno = find_linenum(pos);

	sprintf(buf, "%6d", lno);
	n = strlen(buf);
	column += n;
	for (i = 0;  i < n;  i++)
	{
		STOC(0, buf[i]);
		INC(1);
	}

	/*
	 * Append enough spaces to bring us to the next tab stop.
	 * {{ We could avoid this at the cost of adding some
	 *    complication to the tab stop logic in pappend(). }}
	 */
	do
	{
		STOC(0, ' ');
		INC(1);
		column++;
	} while (column % tabstop);
}

/*
 * Append a character to the line buffer.
 * Expand tabs into spaces, handle underlining, boldfacing, etc.
 * Returns 0 if ok, 1 if couldn't fit in buffer.
 */

#define	NEW_COLUMN(newcol)	if ((newcol) + ((ln_state)?ue_width:0) > sc_width) \
					return (1); else column = (newcol)
	public int
pappend(c)
	int c;
{
	if (c == '\0')
	{
		/*
		 * Terminate any special modes, if necessary.
		 * Append a '\0' to the end of the line.
		 */
		switch (ln_state)
		{
		case LN_UL_X:
			MOVC(0, -1);
			STOC(-1, UE_CHAR);
			INC(1);
			break;
		case LN_BO_X:
			MOVC(0, -1);
			STOC(-1, BE_CHAR);
			INC(1);
			break;
		case LN_UL_XB:
		case LN_UNDERLINE:
			STOC(0, UE_CHAR);
			INC(1);
			break;
		case LN_BO_XB:
		case LN_BOLDFACE:
			STOC(0, BE_CHAR);
			INC(1);
			break;
		}
		ln_state = LN_NORMAL;

		if (column < sc_width || !auto_wrap || ignaw)
		{
			STOC(0, '\n');
			INC(1);
		}
		STOC(0, '\0');
		return (0);
	}

	if (SPACELEFT < 12)
		/*
		 * Almost out of room in the line buffer.
		 * Don't take any chances.
		 * {{ Linebuf is supposed to be big enough that this
		 *    will never happen, but may need to be made
		 *    bigger for wide screens or lots of backspaces. }}
		 */
		return (1);

	if (bs_mode == BS_SPECIAL)
	{
		/*
		 * Advance the state machine.
		 */
		switch (ln_state)
		{
		case LN_NORMAL:
			if (SPACEUSED <= 1 || READC(-1) != '\b')
				break;

			if (c == READC(-2))
				goto enter_boldface;
			if (c == '_' || READC(-2) == '_')
				goto enter_underline;
			INC(-2);
			break;

enter_boldface:
			/*
			 * We have "X\bX" (including the current char).
			 * Switch into boldface mode.
			 */
			if (column + bo_width + be_width + 1 >= sc_width)
				/*
				 * Not enough room left on the screen to
				 * enter and exit boldface mode.
				 */
				return (1);

			if (bo_width > 0 && SPACEUSED > 2 && READC(-3) == ' ')
			{
				/*
				 * Special case for magic cookie terminals:
				 * if the previous char was a space, replace
				 * it with the "enter boldface" sequence.
				 */
				STOC(-3, BO_CHAR);
				column += bo_width-1;
			} else
			{
				MOVC(-1, -2);
				STOC(-2, BO_CHAR);
				column += bo_width;
				INC(1);
			}
			goto ln_bo_xb_case;

enter_underline:
			/*
			 * We have either "_\bX" or "X\b_" (including
			 * the current char).  Switch into underline mode.
			 */
			if (column + ul_width + ue_width + 1 >= sc_width)
				/*
				 * Not enough room left on the screen to
				 * enter and exit underline mode.
				 */
				return (1);

			if (ul_width > 0 && SPACEUSED > 2 && READC(-3) == ' ')
			{
				/*
				 * Special case for magic cookie terminals:
				 * if the previous char was a space, replace
				 * it with the "enter underline" sequence.
				 */
				STOC(-3, UL_CHAR);
				column += ul_width-1;
			} else
			{
				MOVC(-1, -2);
				STOC(-2, UL_CHAR);
				column += ul_width;
				INC(1);
			}
			goto ln_ul_xb_case;
			/*NOTREACHED*/
		case LN_UL_XB:
			/*
			 * Termination of a sequence "_\bX" or "X\b_".
			 */
			if (c != '_' && READC(-2) != '_' && c == READC(-2))
			{
				/*
				 * We seem to have run on from underlining
				 * into boldfacing - this is a nasty fix, but
				 * until this whole routine is rewritten as a
				 * real DFA, ...  well ...
				 */
				MOVC(0, -2);
				STOC(-2, UE_CHAR);
				STOC(-1, BO_CHAR);
				INC(2); /* char & non-existent backspace */
				ln_state = LN_BO_XB;
				goto ln_bo_xb_case;
			}
ln_ul_xb_case:
			if (c == '_')
				c = READC(-2);
			INC(-2);
			ln_state = LN_UNDERLINE;
			break;
		case LN_BO_XB:
			/*
			 * Termination of a sequence "X\bX".
			 */
			if (c != READC(-2) && (c == '_' || READC(-2) == '_'))
			{
				/*
				 * We seem to have run on from
				 * boldfacing into underlining.
				 */
				MOVC(0, -2);
				STOC(-2, BE_CHAR);
				STOC(-1, UL_CHAR);
				INC(2); /* char & non-existent backspace */
				ln_state = LN_UL_XB;
				goto ln_ul_xb_case;
			}
ln_bo_xb_case:
			INC(-2);
			ln_state = LN_BOLDFACE;
			break;
		case LN_UNDERLINE:
			if (column + ue_width + bo_width + 1 + be_width >= sc_width)
				/*
				 * We have just barely enough room to
				 * exit underline mode and handle a possible
				 * underline/boldface run on mixup.
				 */
				return (1);
			ln_state = LN_UL_X;
			break;
		case LN_BOLDFACE:
			if (c == '\b')
			{
				ln_state = LN_BO_XB;
				break;
			}
			if (column + be_width + ul_width + 1 + ue_width >= sc_width)
				/*
				 * We have just barely enough room to
				 * exit underline mode and handle a possible
				 * underline/boldface run on mixup.
				 */
				return (1);
			ln_state = LN_BO_X;
			break;
		case LN_UL_X:
			if (c == '\b')
			{
				ln_state = LN_UL_XB;
			} else
			{
				/*
				 * Exit underline mode.
				 * We have to shuffle the chars a bit
				 * to make this work.
				 */
				MOVC(0, -1);
				STOC(-1, UE_CHAR);
				column += ue_width;
				if (ue_width > 0 && READC(0) == ' ')
				{
					/*
					 * Another special case for magic
					 * cookie terminals: if the next
					 * char is a space, replace it
					 * with the "exit underline" sequence.
					 */
					column--;
				} else
				{
					INC(1);
				}
				ln_state = LN_NORMAL;
			}
			break;
		case LN_BO_X:
			if (c == '\b')
			{
				ln_state = LN_BO_XB;
			} else
			{
				/*
				 * Exit boldface mode.
				 * We have to shuffle the chars a bit
				 * to make this work.
				 */
				MOVC(0, -1);
				STOC(-1, BE_CHAR);
				column += be_width;
				if (be_width > 0 && READC(0) == ' ')
				{
					/*
					 * Another special case for magic
					 * cookie terminals: if the next
					 * char is a space, replace it
					 * with the "exit boldface" sequence.
					 */
					column--;
				} else
				{
					INC(1);
				}
				ln_state = LN_NORMAL;
			}
			break;
		}
	}

	if (c == '\t')
	{
		/*
		 * Expand a tab into spaces.
		 */
		do
		{
			STOC(0, ' ');
			INC(1);
			NEW_COLUMN(column+1);
		} while ((column % tabstop) != 0);
		return (0);
	}

	if (c == '\b')
	{
		if (bs_mode == BS_CONTROL)
		{
			/*
			 * Treat backspace as a control char: output "^H".
			 */
			NEW_COLUMN(column+2);
			STOC(0, ('H' | CARATBIT));
			INC(1);
		} else
		{
			/*
			 * Output a real backspace.
			 */
			column--;
			STOC(0, '\b');
			INC(1);
		}
		return (0);
        }

        if (c == '\r')
          return(0);

	if (control_char(c) && !sendctl)
	{
		/*
		 * Put a "^X" into the buffer.
		 * The CARATBIT bit is used to tell put_line() to prefix
		 * the char with a ^.  We don't actually put the ^
		 * in the buffer because we sometimes need to move
		 * chars around, and such movement might separate
		 * the ^ from its following character.
		 */
		NEW_COLUMN(column+2);
		STOC(0, (carat_char(c) | CARATBIT));
		INC(1);
		return (0);
	}

	/*
	 * Ordinary character.  Just put it in the buffer.
	 */
	NEW_COLUMN(column+1);
	STOC(0, c);
	INC(1);
	return (0);
}

/*
 * Get a character from the current line.
 */
	public int
gline(i)
	int i;
{
	if (is_null_line)
	{
		/*
		 * If there is no current line, we pretend the line is
		 * either "~" or "", depending on the "twiddle" flag.
		 */
		if (twiddle)
			switch (i)
			{
			case 0:  return ('~');
			case 1:  return ('\n');
			}
		return (0);
	}

	return ((linebuf[i]&0377) | ((xlinebuf[i]&0377)<<8));
}

/*
 * Indicate that there is no current line.
 */
	public void
null_line()
{
	is_null_line = 1;
}

/*
 * Analogous to forw_line(), but deals with "raw lines":
 * lines which are not split for screen width.
 * {{ This is supposed to be more efficient than forw_line(). }}
 */
	public POSITION
forw_raw_line(curr_pos, linep)
	POSITION curr_pos;
	char **linep;
{
	register char *p;
	register int c;
	POSITION new_pos;

	if (curr_pos == NULL_POSITION || ch_seek(curr_pos) ||
		(c = ch_forw_get()) == EOI)
		return (NULL_POSITION);

	p = linebuf;

	for (;;)
	{
		if (c == '\n' || c == EOI)
		{
			new_pos = ch_tell();
			break;
		}
		if (p >= &linebuf[sizeof(linebuf)-1])
		{
			/*
			 * Overflowed the input buffer.
			 * Pretend the line ended here.
			 * {{ The line buffer is supposed to be big
			 *    enough that this never happens. }}
			 */
			new_pos = ch_tell() - 1;
			break;
		}
		*p++ = c;
		c = ch_forw_get();
	}
	*p = '\0';
	if (linep != NULL)
		*linep = linebuf;
	return (new_pos);
}

/*
 * Analogous to back_line(), but deals with "raw lines".
 * {{ This is supposed to be more efficient than back_line(). }}
 */
	public POSITION
back_raw_line(curr_pos, linep)
	POSITION curr_pos;
	char **linep;
{
	register char *p;
	register int c;
	POSITION new_pos;

	if (curr_pos == NULL_POSITION || curr_pos <= (POSITION)0 ||
		ch_seek(curr_pos-1))
		return (NULL_POSITION);

	p = &linebuf[sizeof(linebuf)];
	*--p = '\0';

	for (;;)
	{
		c = ch_back_get();
		if (c == '\n')
		{
			/*
			 * This is the newline ending the previous line.
			 * We have hit the beginning of the line.
			 */
			new_pos = ch_tell() + 1;
			break;
		}
		if (c == EOI)
		{
			/*
			 * We have hit the beginning of the file.
			 * This must be the first line in the file.
			 * This must, of course, be the beginning of the line.
			 */
			new_pos = (POSITION)0;
			break;
		}
		if (p <= linebuf)
		{
			/*
			 * Overflowed the input buffer.
			 * Pretend the line ended here.
			 */
			new_pos = ch_tell() + 1;
			break;
		}
		*--p = c;
	}
	if (linep != NULL)
		*linep = p;
	return (new_pos);
}
