/*
 * Handling functions for command line options.
 *
 * Most options are handled by the generic code in option.c.
 * But all string options, and a few non-string options, require
 * special handling specific to the particular option.
 * This special processing is done by the "handling functions" in this file.
 *
 * Each handling function is passed a "type" and, if it is a string
 * option, the string which should be "assigned" to the option.
 * The type may be one of:
 *	INIT	The option is being initialized from the command line.
 *	TOGGLE	The option is being changed from within the program.
 *	QUERY	The setting of the option is merely being queried.
 */

#include "less.h"
#include "option.h"

extern int nbufs;
extern int ispipe;
extern int cbufs;
extern int pr_type;
extern int seven_bit;
extern int nohelp;
extern char *prproto[];
extern char *eqproto;
#if LOGFILE
extern char *namelogfile;
extern int force_logfile;
extern int logfile;
#endif
#if TAGS
public int tagoption = 0;
extern char *tagfile;
extern char *tagpattern;
#endif


#if LOGFILE
/*
 * Handler for -l option.
 */
	public void
opt_l(type, s)
	int type;
	char *s;
{
	char *m;

	switch (type)
	{
	case INIT:
		namelogfile = s;
		break;
	case TOGGLE:
		if (!ispipe)
		{
			error("Input is not a pipe");
			return;
		}
		if (logfile >= 0)
		{
			error("Log file is already in use");
			return;
		}
		namelogfile = save(skipsp(s));
		use_logfile();
		sync_logfile();
		break;
	case QUERY:
		if (logfile < 0)
			error("No log file");
		else
		{
			static char lf[] = "log file \"";
			m = ecalloc(sizeof(lf) + strlen(namelogfile) + 2, sizeof(char));
			strcpy(m, lf);
			strcat(m, namelogfile);
			strcat(m, "\"");
			error(m);
			free(m);
		}
		break;
	}
}

/*
 * Handler for -L option.
 */
	public void
opt__L(type, s)
	int type;
	char *s;
{
	force_logfile = 1;
	opt_l(type, s);
}
#endif

#if USERFILE
	public void
opt_k(type, s)
	int type;
	char *s;
{
	char *message;
	static char MSG[] = "Cannot use lesskey file: ";

	switch (type)
	{
	case INIT:
		if (add_cmdtable(s) == 0)
			return;
		message = (char *) ecalloc(strlen(s) + sizeof(MSG) + 1,
						sizeof(char));
		strcpy(message, MSG);
		strcat(message, s);
		error(message);
		free(message);
		break;
	case QUERY:
	case TOGGLE:
		error("Cannot query -k option");
		break;
	}
}
#endif

#if TAGS
/*
 * Handler for -t option.
 */
	public void
opt_t(type, s)
	int type;
	char *s;
{
	switch (type)
	{
	case INIT:
		tagoption = 1;
		findtag(s);
		break;
	case TOGGLE:
		findtag(skipsp(s));
		if (tagfile != NULL)
		{
			edit(tagfile);
			(void) tagsearch();
		}
		break;
	case QUERY:
		error("Tag is required after -t");
		break;
	}
}
#endif

/*
 * Handler for -P option.
 */
	public void
opt_P(type, s)
	int type;
	register char *s;
{
	register char **proto;

	switch (type)
	{
	case INIT:
	case TOGGLE:
		/*
		 * Figure out which prototype string should be changed.
		 */
		switch (*s)
		{
		case 'm':  proto = &prproto[PR_MEDIUM];	s++;	break;
		case 'M':  proto = &prproto[PR_LONG];	s++;	break;
		case '=':  proto = &eqproto;		s++;	break;
		default:   proto = &prproto[pr_type];		break;
		}
		free(*proto);
		*proto = save(s);
		break;
	case QUERY:
		error(prproto[pr_type]);
		break;
	}
}

/*
 * Handler for the -b option.
 */
	public void
opt_b(type)
	int type;
{
	switch (type)
	{
	case TOGGLE:
		/*
		 * Allocate the new number of buffers.
		 */
		ch_init(cbufs, 1);
		break;
	case QUERY:
	case INIT:
		break;
	}
}

/*
 * Handler for the -g option.
 */
	public void
opt_g(type)
	int type;
{
	switch (type)
	{
	case TOGGLE:
		/*
		 * Flush the buffers, since the buffered data may
		 * be bad (if we are switching from 7 bits to 8 bits,
		 * the eighth bit has been stripped from the buffered data).
		 */
		if (ispipe)
		{
			error("Can't change -g on a pipe");
			seven_bit = !seven_bit;
			break;
		}
		ch_init(0, 0);
		break;
	case INIT:
	case QUERY:
		break;
	}
}

/*
 * "-?" means display a help message.
 * If from the command line, exit immediately.
 */
	public void
opt_query(type)
	int type;
{
	if (nohelp)
		return;
	switch (type)
	{
	case QUERY:
	case TOGGLE:
		error("Use \"h\" for help");
		break;
	case INIT:
		raw_mode(1);
		init();
		help();
		quit();
		/*NOTREACHED*/
	}
}
