/*
 * Operating system dependent routines.
 *
 * Most of the stuff in here is based on Unix, but an attempt
 * has been made to make things work on other operating systems.
 * This will sometimes result in a loss of functionality, unless
 * someone rewrites code specifically for the new operating system.
 *
 * The makefile provides defines to decide whether various
 * Unix features are present.
 */

#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include "less.h"

#ifdef MSDOS
#include <process.h>
extern int _osmode;
#define DOS_MODE 0
#endif

extern char *getenv();

public int reading;

extern int screen_trashed;
extern int force_open;
extern char *current_file;
extern char *previous_file;

static jmp_buf read_label;

/*
 * Pass the specified command to a shell to be executed.
 * Like plain "system()", but handles resetting terminal modes, etc.
 */
	public void
lsystem(cmd)
	char *cmd;
{
	register int inp;
	register char *shell;
	register char *p;

	/*
	 * Print the command which is to be executed,
	 * unless the command starts with a "-".
	 */
	if (cmd[0] == '-')
		cmd++;
	else
	{
		lower_left();
		clear_eol();
		putstr("!");
		putstr(cmd);
		putstr("\n");
	}

	/*
	 * De-initialize the terminal and take out of raw mode.
	 */
	deinit();
	flush();
	raw_mode(0);

	/*
	 * Restore signals to their defaults.
	 */
	init_signals(0);

	/*
	 * Force standard input to be the terminal, "/dev/tty",
	 * even if less's standard input is coming from a pipe.
	 */
	inp = dup(0);
        close(0);

#ifdef MSDOS
        if (open("con", 0) < 0)
#else
        if (open("/dev/tty", 0) < 0)
#endif
		dup(inp);

#ifdef MSDOS
        if ( *cmd == 0 )
        {
          if ( (p = getenv("COMSPEC")) == NULL )
            p = (_osmode == DOS_MODE) ? "COMMAND" : "CMD";

          spawnlp(P_WAIT, p, p, NULL);
        }
        else
          system(cmd);
#else
	/*
	 * Pass the command to the system to be executed.
	 * If we have a SHELL environment variable, use
	 * <$SHELL -c "command"> instead of just <command>.
	 * If the command is empty, just invoke a shell.
	 */
	p = NULL;
	if ((shell = getenv("SHELL")) != NULL && *shell != '\0')
	{
		if (*cmd == '\0')
			p = save(shell);
		else
		{
			p = ecalloc(strlen(shell) + strlen(cmd) + 7, sizeof(char));
			sprintf(p, "%s -c \"%s\"", shell, cmd);
		}
	}
	if (p == NULL)
                p = save("sh");

	system(p);
	free(p);
#endif

	/*
	 * Restore standard input, reset signals, raw mode, etc.
	 */
	close(0);
	dup(inp);
	close(inp);

	init_signals(1);
	raw_mode(1);
	init();
	screen_trashed = 1;
#if defined(SIGWINCH) || defined(SIGWIND)
	/*
	 * Since we were ignoring window change signals while we executed
	 * the system command, we must assume the window changed.
	 * Warning: this leaves a signal pending (in "sigs"),
	 * so psignals() should be called soon after lsystem().
	 */
	winch();
#endif
}

/*
 * Like read() system call, but is deliberately interruptible.
 * A call to intread() from a signal handler will interrupt
 * any pending iread().
 */
	public int
iread(fd, buf, len)
	int fd;
	char *buf;
	int len;
{
	register int n;

	if (setjmp(read_label))
	{
		/*
		 * We jumped here from intread.
		 */
		reading = 0;
#if SIGSETMASK
		sigsetmask(0);
#endif
		return (READ_INTR);
	}

	flush();
	reading = 1;
	n = read(fd, buf, len);
	reading = 0;
	if (n < 0)
		return (-1);
	return (n);
}

	public void
intread()
{
	longjmp(read_label, 1);
}

#if GET_TIME
	public long
get_time()
{
	long t;

	time(&t);
	return (t);
}
#endif

/*
 * Expand a string, substituting any "%" with the current filename,
 * and any "#" with the previous filename.
 */
	public char *
fexpand(s)
	char *s;
{
	register char *fr, *to;
	register int n;
	register char *e;

	/*
	 * Make one pass to see how big a buffer we
	 * need to allocate for the expanded string.
	 */
	n = 0;
	for (fr = s;  *fr != '\0';  fr++)
	{
		if (*fr == '%')
			n += strlen(current_file);
		else if (*fr == '#')
		{
			if (previous_file == NULL)
			{
				error("No previous file");
				return (NULL);
			}
			n += strlen(previous_file);
		} else
			n++;
	}

	e = ecalloc(n+1, sizeof(char));

	/*
	 * Now copy the string, expanding any "%" or "#".
	 */
	to = e;
	for (fr = s;  *fr != '\0';  fr++)
	{
		if (*fr == '%')
		{
			strcpy(to, current_file);
			to += strlen(to);
		} else if (*fr == '#')
		{
			strcpy(to, previous_file);
			to += strlen(to);
		} else
			*to++ = *fr;
	}
	*to = '\0';
	return (e);
}

	static POSITION
seek_filesize(f)
	int f;
{
	offset_t spos;

	spos = lseek(f, (offset_t)0, 2);
	if (spos == BAD_LSEEK)
		return (NULL_POSITION);
	return ((POSITION) spos);
}

/*
 * Expand a filename, substituting any environment variables, etc.
 * The implementation of this is necessarily very operating system
 * dependent.  This implementation is unabashedly only for Unix systems.
 */
#if GLOB

FILE *popen();

	public char *
glob(filename)
	char *filename;
{
	FILE *f;
	char *p;
	int ch;
	char *cmd;
	char *gfilename;

	filename = fexpand(filename);
	if (filename == NULL)
		return (NULL);

	/*
	 * We get the shell to expand the filename for us by passing
	 * an "echo" command to the shell and reading its output.
	 */
	p = getenv("SHELL");
	if (p == NULL || *p == '\0')
	{
		/*
		 * Read the output of <echo filename>.
		 */
		cmd = ecalloc(strlen(filename)+6, sizeof(char));
		sprintf(cmd, "echo %s", filename);
	} else
	{
		/*
		 * Read the output of <$SHELL -c "echo filename">.
		 */
		cmd = ecalloc(strlen(p)+strlen(filename)+12, sizeof(char));
		sprintf(cmd, "%s -c \"echo %s\"", p, filename);
	}

	f = popen(cmd, "r");
	free(cmd);
	if (f == NULL)
		return (filename);
	free(filename);

	gfilename = ecalloc(FILENAME, sizeof(char));
	for (p = gfilename;  p < &gfilename[FILENAME-1];  p++)
	{
		if ((ch = getc(f)) == '\n' || ch == EOF)
			break;
		*p = ch;
	}
	*p = '\0';
	pclose(f);
	return (gfilename);
}

#else

	public char *
glob(filename)
	char *filename;
{
	return (fexpand(filename));
}

#endif


/*
 * Returns NULL if the file can be opened and
 * is an ordinary file, otherwise an error message
 * (if it cannot be opened or is a directory, etc.)
 */

#if STAT

#include <sys/types.h>
#include <sys/stat.h>

	public char *
bad_file(filename)
	char *filename;
{
	register char *m;
	struct stat statbuf;

	if (stat(filename, &statbuf) < 0)
		return (errno_message(filename));

	if (force_open)
		return (NULL);

	if ((statbuf.st_mode & S_IFMT) == S_IFDIR)
	{
		static char is_dir[] = " is a directory";
		m = ecalloc(strlen(filename) + sizeof(is_dir), sizeof(char));
		strcpy(m, filename);
		strcat(m, is_dir);
		return (m);
	}
	if ((statbuf.st_mode & S_IFMT) != S_IFREG)
	{
		static char not_reg[] = " is not a regular file";
		m = ecalloc(strlen(filename) + sizeof(not_reg), sizeof(char));
		strcpy(m, filename);
		strcat(m, not_reg);
		return (m);
	}
	return (NULL);
}

	public POSITION
filesize(f)
	int f;
{
	struct stat statbuf;

	if (fstat(f, &statbuf) < 0)
		return (seek_filesize(f));
	return ((POSITION) statbuf.st_size);
}

#else

	public char *
bad_file(filename)
	char *filename;
{
	return (NULL);
}

	public POSITION
filesize(f)
	int f;
{
	return (seek_filesize(f));
}

#endif

/*
 * errno_message: Return an error message based on the value of "errno".
 */

#if PERROR

extern char *sys_errlist[];
extern int sys_nerr;
extern int errno;

	public char *
errno_message(filename)
	char *filename;
{
	register char *p;
	register char *m;
	char msg[16];

	if (errno < sys_nerr)
		p = sys_errlist[errno];
	else
	{
		sprintf(msg, "Error %d", errno);
		p = msg;
	}
	m = ecalloc(strlen(filename) + strlen(p) + 3, sizeof(char));
	sprintf(m, "%s: %s", filename, p);
	return (m);
}

#else

	public char *
errno_message(filename)
	char *filename;
{
	register char *m;
	static char msg[] = ": cannot open";

	m = ecalloc(strlen(filename) + sizeof(msg), sizeof(char));
	strcpy(m, filename);
	strcat(m, msg);
	return (m);
}

#endif
