/*
 * More primitives for displaying the file on the screen.
 */

#include "less.h"
#include "position.h"

extern int sigs;
extern int how_search;
extern int top_scroll;
extern int back_scroll;
extern int caseless;
extern int linenums;
extern int sc_height;
extern HANDLE curr_handle;

/*
 * Jump to a specified position in the file.
 * The position need not be the first character in a line.
 */
	static void
jump_line_loc(pos)
	POSITION pos;
{
	int c;

	if (ch_seek(pos) == 0)
	{
		/*
		 * Back up to the beginning of the line.
		 */
		while ((c = ch_back_get()) != '\n' && c != EOI)
			;
		if (c == '\n')
			(void) ch_forw_get();
		pos = ch_tell();
	}
	jump_loc(pos);
}

/*
 * Jump to a specified percentage into the file.
 * This is a poor compensation for not being able to
 * quickly jump to a specific line number.
 */
	public void
jump_percent(percent)
	int percent;
{
	POSITION pos, len;

	/*
	 * Determine the position in the file
	 * (the specified percentage of the file's length).
	 */
	if ((len = ch_length()) == NULL_POSITION)
	{
		error("Don't know length of file");
		return;
	}
	pos = (percent * len) / 100;

	jump_line_loc(pos);
}

/*
 * The table of marks.
 * A mark is simply a position in a file and the handle of the file.
 */
#define	NMARKS		(27)		/* 26 for a-z plus one for quote */
#define	LASTMARK	(NMARKS-1)	/* For quote */

static struct mark {
	HANDLE m_handle;
	POSITION m_pos;
} marks[NMARKS];

/*
 * Initialize the mark table to show no marks are set.
 */
	public void
init_mark()
{
	int i;

	for (i = 0;  i < NMARKS;  i++)
		marks[i].m_pos = NULL_POSITION;
}

/*
 * See if a mark letter is valid (between a and z).
 */
	static int
badmark(c)
	int c;
{
	if (c < 'a' || c > 'z')
	{
		error("Choose a letter between 'a' and 'z'");
		return (1);
	}
	return (0);
}

/*
 * Set a mark.
 */
	public void
setmark(c)
	int c;
{
	if (badmark(c))
		return;
	c -= 'a';
	marks[c].m_pos = position(TOP);
	marks[c].m_handle = curr_handle;
}

/*
 * Set the LASTMARK (the mark named by the apostrophe).
 */
	public void
lastmark()
{
	POSITION pos;

	pos = position(TOP);
	if (pos == NULL_POSITION)
		return;
	marks[LASTMARK].m_pos = pos;
	marks[LASTMARK].m_handle = curr_handle;
}

/*
 * Go to a previously set mark.
 */
	public void
gomark(c)
	int c;
{
	POSITION pos;
	HANDLE handle;

	if (c == '\'')
		c = LASTMARK;
	else if (badmark(c))
		return;
	else
		c -= 'a';

	pos = marks[c].m_pos;
	handle = marks[c].m_handle;
	if (pos == NULL_POSITION)
	{
		error("Mark not set");
		return;
	}

	if (handle != curr_handle)
		/*
		 * Not in the current file; edit the correct file.
		 */
		edit(get_filename(handle));

	jump_loc(pos);
}

/*
 * Get the backwards scroll limit.
 * Must call this function instead of just using the value of
 * back_scroll, because the default case depends on sc_height and
 * top_scroll, as well as back_scroll.
 */
	public int
get_back_scroll()
{
	if (back_scroll >= 0)
		return (back_scroll);
	if (top_scroll)
		return (sc_height - 2);
	return (10000); /* infinity */
}

/*
 * Try to match the n-th bracket of the specified type
 * which appears in the top displayed line.
 * brac may be '\0' to mean look for any bracket.
 * "Bracket" refers to any of the pairs: { }, [ ], or ( ).
 */
	public void
match_brac(brac, n)
	register int brac;
	int n;
{
	register int c;
	register int nest;
	int obrac, cbrac;
	int forwdir;
	POSITION pos;
	int (*chget)();

	extern int ch_forw_get(), ch_back_get();

	pos = position(TOP);
	if (pos == NULL_POSITION || ch_seek(pos))
	{
		error("Nothing in top line");
		return;
	}

	/*
	 * Look thru the first line to find the type of bracket to match.
	 */
	for (;;)
	{
		if ((c = ch_forw_get()) == '\n' || c == EOI)
		{
			error("No bracket in top line");
			return;
		}
		if (brac != '\0' && brac != c)
			/*
			 * This is not the specified bracket character.
			 */
			continue;

		switch (c)
		{
		default:  continue;
		case '{':   obrac = '{';  cbrac = '}';  forwdir = 1;  break;
		case '}':   obrac = '}';  cbrac = '{';  forwdir = 0;  break;
		case '[':   obrac = '[';  cbrac = ']';  forwdir = 1;  break;
		case ']':   obrac = ']';  cbrac = '[';  forwdir = 0;  break;
		case '(':   obrac = '(';  cbrac = ')';  forwdir = 1;  break;
		case ')':   obrac = ')';  cbrac = '(';  forwdir = 0;  break;
		}
		/*
		 * See if we have the n-th bracket in the line.
		 */
		if (--n <= 0)
			break;
	}

	if (!forwdir)
	{
		/*
		 * Position the file just "after" the open bracket.
		 * That is, if searching backwards, skip back over
		 * the open bracket now.
		 */
		(void) ch_back_get();
	}

	/*
	 * Search the file for the matching bracket.
	 */
	chget = (forwdir) ? ch_forw_get : ch_back_get;
	nest = 0;
	while ((c = (*chget)()) != EOI)
	{
		if (c == obrac)
			nest++;
		else if (c == cbrac && --nest < 0)
		{
			jump_line_loc(ch_tell());
			return;
		}
	}
	error("No matching bracket");
}

/*
 * Search for the n-th occurrence of a specified pattern,
 * either forward or backward.
 */
	public void
search(search_type, pattern, n)
	int search_type;
	char *pattern;
	register int n;
{
	POSITION pos, linepos;
	register char *p;
	register char *q;
	register int goforw;
	register int wantmatch;
	char *line;
	int linenum;
	int linematch;
#if RECOMP
	char *re_comp();
	char *errmsg;
#else
#if REGCMP
	char *regcmp();
	static char *cpattern = NULL;
#else
	static char lpbuf[100];
	static char *last_pattern = NULL;
#endif
#endif

	/*
	 * Extract flags and type of search.
	 */
	wantmatch = !(search_type & SRCH_NOMATCH);
	search_type = SRCH_TYPE(search_type);

	if (caseless && pattern != NULL)
	{
		/*
		 * For a caseless search, convert any uppercase
		 * in the pattern to lowercase.
		 */
		for (p = pattern;  *p != '\0';  p++)
			if (*p >= 'A' && *p <= 'Z')
				*p += 'a' - 'A';
	}
#if RECOMP

	/*
	 * (re_comp handles a null pattern internally,
	 *  so there is no need to check for a null pattern here.)
	 */
	if ((errmsg = re_comp(pattern)) != NULL)
	{
		error(errmsg);
		return;
	}
#else
#if REGCMP
	if (pattern == NULL || *pattern == '\0')
	{
		/*
		 * A null pattern means use the previous pattern.
		 * The compiled previous pattern is in cpattern, so just use it.
		 */
		if (cpattern == NULL)
		{
			error("No previous regular expression");
			return;
		}
	} else
	{
		/*
		 * Otherwise compile the given pattern.
		 */
		char *s;
		if ((s = regcmp(pattern, 0)) == NULL)
		{
			error("Invalid pattern");
			return;
		}
		if (cpattern != NULL)
			free(cpattern);
		cpattern = s;
	}
#else
	if (pattern == NULL || *pattern == '\0')
	{
		/*
		 * Null pattern means use the previous pattern.
		 */
		if (last_pattern == NULL)
		{
			error("No previous regular expression");
			return;
		}
		pattern = last_pattern;
	} else
	{
		strcpy(lpbuf, pattern);
		last_pattern = lpbuf;
	}
#endif
#endif

	/*
	 * Figure out where to start the search.
	 */

	if (search_type == SRCH_FILE)
	{
		/*
		 * User wants to start searching at beginning of file.
		 * {{ Use ch_beg_seek() in case we can't seek to 0? }}
		 */
		pos = (POSITION)0;
		goforw = 1;
	} else if (position(TOP) == NULL_POSITION)
	{
		/*
		 * Nothing is currently displayed.
		 * Start at the beginning of the file.
		 * (This case is mainly for first_cmd searches,
		 * for example, "+/xyz" on the command line.)
		 */
		pos = (POSITION)0;
		goforw = 1;
	} else if (search_type == SRCH_BACK)
	{
		/*
		 * Backward search: start just before the top line
		 * displayed on the screen.
		 */
		pos = position(TOP);
		goforw = 0;
	} else if (how_search == 0)
	{
		/*
		 * Start at the second real line displayed on the screen.
		 */
		pos = position(TOP);
		do
			pos = forw_raw_line(pos, (char **)NULL);
		while (pos < position(TOP_PLUS_ONE));
		goforw = 1;
	} else if (how_search == 1)
	{
		/*
		 * Start just after the bottom line displayed on the screen.
		 */
		pos = position(BOTTOM_PLUS_ONE);
		goforw = 1;
	} else
	{
		/*
		 * Start at the second screen line displayed on the screen.
		 */
		pos = position(TOP_PLUS_ONE);
		goforw = 1;
	}

	if (pos == NULL_POSITION)
	{
		/*
		 * Can't find anyplace to start searching from.
		 */
		error("Nothing to search");
		return;
	}

	linenum = find_linenum(pos);
	for (;;)
	{
		/*
		 * Get lines until we find a matching one or
		 * until we hit end-of-file (or beginning-of-file
		 * if we're going backwards).
		 */
		if (sigs)
			/*
			 * A signal aborts the search.
			 */
			return;

		if (goforw)
		{
			/*
			 * Read the next line, and save the
			 * starting position of that line in linepos.
			 */
			linepos = pos;
			pos = forw_raw_line(pos, &line);
			if (linenum != 0)
				linenum++;
		} else
		{
			/*
			 * Read the previous line and save the
			 * starting position of that line in linepos.
			 */
			pos = back_raw_line(pos, &line);
			linepos = pos;
			if (linenum != 0)
				linenum--;
		}

		if (pos == NULL_POSITION)
		{
			/*
			 * We hit EOF/BOF without a match.
			 */
			error("Pattern not found");
			return;
		}

		/*
		 * If we're using line numbers, we might as well
		 * remember the information we have now (the position
		 * and line number of the current line).
		 */
		if (linenums)
			add_lnum(linenum, pos);

		if (caseless)
		{
			/*
			 * If this is a caseless search, convert
			 * uppercase in the input line to lowercase.
			 * While we're at it, remove any backspaces
			 * along with the preceding char.
			 * This allows us to match text which is
			 * underlined or overstruck.
			 */
			for (p = q = line;  *p != '\0';  p++, q++)
			{
				if (*p >= 'A' && *p <= 'Z')
					/* Convert uppercase to lowercase. */
					*q = *p + 'a' - 'A';
				else if (q > line && *p == '\b')
					/* Delete BS and preceding char. */
					q -= 2;
				else
					/* Otherwise, just copy. */
					*q = *p;
			}
		}

		/*
		 * Test the next line to see if we have a match.
		 * This is done in a variety of ways, depending
		 * on what pattern matching functions are available.
		 */
#if REGCMP
		linematch = (regex(cpattern, line) != NULL);
#else
#if RECOMP
		linematch = (re_exec(line) == 1);
#else
		linematch = match(pattern, line);
#endif
#endif
		/*
		 * We are successful if wantmatch and linematch are
		 * both true (want a match and got it),
		 * or both false (want a non-match and got it).
		 */
		if (((wantmatch && linematch) || (!wantmatch && !linematch)) &&
		      --n <= 0)
			/*
			 * Found the line.
			 */
			break;
	}

	jump_loc(linepos);
}

#if (!REGCMP) && (!RECOMP)
/*
 * We have neither regcmp() nor re_comp().
 * We use this function to do simple pattern matching.
 * It supports no metacharacters like *, etc.
 */
	static int
match(pattern, buf)
	char *pattern, *buf;
{
	register char *pp, *lp;

	for ( ;  *buf != '\0';  buf++)
	{
		for (pp = pattern, lp = buf;  *pp == *lp;  pp++, lp++)
			if (*pp == '\0' || *lp == '\0')
				break;
		if (*pp == '\0')
			return (1);
	}
	return (0);
}
#endif
